from abc import ABC, abstractmethod
from collections.abc import MutableMapping
from functools import singledispatch
from typing import Mapping, Optional, Tuple
import warnings

import numpy as np
import pandas as pd
from scipy.sparse import spmatrix

from ..utils import deprecated
from .views import asview, ViewArgs


@singledispatch
def _subset(mat, subset_idx):
    return mat[subset_idx]

@_subset.register(pd.DataFrame)
def _subset_df(df: pd.DataFrame, subset_idx):
    return df.iloc[subset_idx]


class AlignedMapping(MutableMapping, ABC):
    """
    Attributes
    ----------
    axes: Tuple[int]
        Which axes is this aligned along?
    _view_class:
        The view class for this aligned mapping.
    _actual_class
        The actual class (which has it's own data) for this aligned mapping.
    """

    def __repr__(self):
        return "{} with keys: {}".format(
            self.__class__.__name__, ', '.join(self.keys())
        )

    def _validate_value(self, val, key) -> None:
        """Raises an error if value is invalid"""
        for i, axis in enumerate(self.axes):
            if self.parent.shape[axis] != val.shape[i]:
                right_shape = tuple(self.parent.shape[a] for a in self.axes)
                raise ValueError(
                    "Value passed for key '{key}' is of incorrect shape. Values of"
                    " {attrname} must match dimensions {axes} of parent. value had"
                    " shape {wrong_shape} while it should have had {right_shape}."
                    .format(
                        key=key, attrname=self.attrname, axes=self.axes, wrong_shape=val.shape, right_shape=right_shape)
                )
        try:  # TODO: Handle objects with indices
            # Could probably also re-order index if it's contained
            if not (val.index == self.dim_names).all():
                raise IndexError()  # Maybe not index error
        except AttributeError:
            pass
        # TODO: Modify this as soon as writing dataframes works
        if not isinstance(val, (np.ndarray, spmatrix)):
            warnings.warn(
                f"AnnData does not currently support writing or reading of "
                f"'{type(val).__name__}' objects in {self.attrname} for either"
                f" hdf5 or zarr formats.",
                stacklevel=2
            )

    @property
    @abstractmethod
    def attrname(self) -> str:
        """What attr for the AnnData is this?"""
        pass

    @property
    @abstractmethod
    def axes(self) -> Tuple[int, ...]:
        """Which axes of the parent is this aligned to?"""
        pass

    @property
    @abstractmethod
    def isview(self) -> bool:
        pass

    @property
    def parent(self) -> "AnnData":
        return self._parent

    def copy(self):
        d = self._actual_class(self.parent, self._axis)
        for k, v in self.items():
            d[k] = v.copy()
        return d

    def _view(self, parent, subset_idx):
        """Returns a subset copy-on-write view of the object."""
        return self._view_class(self, parent, subset_idx)

    @deprecated("dict(obj)")
    def as_dict(self) -> dict:
        return dict(self)


class AlignedViewMixin:
    """
    Attributes
    ----------
    parent
        Reference to parent AnnData view
    attrname
        What attribute in the parent is this?
    parent_mapping
        The object this is a view of.
    """
    isview = True

    def __getitem__(self, key):
        return _subset(
            asview(
                self.parent_mapping[key],
                ViewArgs(self.parent, self.attrname, (key,))
            ),
            self.subset_idx
        )

    def __setitem__(self, key, value):
        self._validate_value(value, key)  # Validate before mutating
        adata = self.parent.copy()
        new_mapping = getattr(adata, self.attrname)
        new_mapping[key] = value
        self.parent._init_as_actual(adata)

    def __delitem__(self, key):
        self[key]  # Make sure it exists before bothering with a copy
        adata = self.parent.copy()
        new_mapping = getattr(adata, self.attrname)
        del new_mapping[key]
        self.parent._init_as_actual(adata)

    def __contains__(self, key):
        return key in self.parent_mapping

    def __iter__(self):
        return iter(self.parent_mapping)

    def __len__(self):
        return len(self.parent_mapping)


class AlignedActualMixin:
    """
    Attributes
    _data: dict
        Underlying mapping to the data
    """
    isview = False

    def __getitem__(self, key):
        return self._data[key]

    def __setitem__(self, key, value):
        self._validate_value(value, key)
        self._data[key] = value

    def __contains__(self, k):
        return k in self._data

    def __delitem__(self, key):
        del self._data[key]

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)


class AxisArraysBase(AlignedMapping):
    """
    Mapping of key: array-like, where array-like is aligned to an axis of parent AnnData.
    """
    _dimnames = ("obs", "var")

    @property
    def attrname(self):
        return "{}m".format(self.dim)

    @property
    def axes(self):
        """Axes of the parent this is aligned to"""
        return (self._axis,)

    @property
    def dim(self):
        """Name of the dimension this aligned to."""
        return self._dimnames[self._axis]

    def flipped(self):
        """Transpose."""
        new = self.copy()
        new.dimension = abs(self._axis - 1)
        return new

    def to_df(self) -> pd.DataFrame:
        """Convert to pandas dataframe."""
        df = pd.DataFrame(index=self.dim_names)
        for key in self.keys():
            value = self[key]
            for icolumn, column in enumerate(value.T):
                df['{}{}'.format(key, icolumn + 1)] = column
        return df


class AxisArrays(AlignedActualMixin, AxisArraysBase):
    def __init__(self, parent: "AnnData", axis: int, vals: Optional[Mapping] = None):
        self._parent = parent
        if axis not in (0, 1):
            raise ValueError()
        self._axis = axis
        self.dim_names = (parent.obs_names, parent.var_names)[self._axis]
        self._data = dict()
        if vals is not None:
            self.update(vals)


class AxisArraysView(AlignedViewMixin, AxisArraysBase):
    def __init__(self, parent_mapping, parent_view, subset_idx):
        self.parent_mapping = parent_mapping
        self._parent = parent_view
        self.subset_idx = subset_idx
        self._axis = parent_mapping._axis
        self.dim_names = parent_mapping.dim_names[subset_idx]


AxisArraysBase._view_class = AxisArraysView
AxisArraysBase._actual_class = AxisArrays


class LayersBase(AlignedMapping):
    """
    Mapping of key: array-like, where array-like is aligned to both axes of the
    parent anndata.
    """
    attrname = "layers"
    axes = (0, 1)


class Layers(AlignedActualMixin, LayersBase):
    def __init__(self, parent: "AnnData", vals: Optional[Mapping] = None):
        self._parent = parent
        self._data = dict()
        if vals is not None:
            self.update(vals)


class LayersView(AlignedViewMixin, LayersBase):
    def __init__(self, parent_mapping: LayersBase, parent_view: "AnnData", subset_idx):
        self.parent_mapping = parent_mapping
        self._parent = parent_view
        self.subset_idx = subset_idx


LayersBase._view_class = LayersView
LayersBase._actual_class = Layers


class PairwiseArraysBase(AlignedMapping):
    """
    Mapping of key: array-like, where both axes of array-like are aligned to
    one axis of the parent anndata.
    """
    _dimnames = ("obs", "var")

    @property
    def attrname(self):
        return "{}p".format(self.dim)

    @property
    def axes(self):
        """Axes of the parent this is aligned to"""
        return (self._axis, self._axis)

    @property
    def dim(self):
        """Name of the dimension this aligned to."""
        return self._dimnames[self._axis]


# class PairwiseArrays(AlignedActualMixin, PairwiseArraysBase):
#     def __init__(self, parent: "AnnData", axis: int, vals: Optional[Mapping] = None):
#         self._parent = parent
#         if axis not in (0, 1):
#             raise ValueError()
#         self._axis = axis
#         self.dim_names = (parent.obs_names, parent.var_names)[self._axis]
#         self._data = dict()
#         if vals is not None:
#             self.update(vals)


# class PairwiseArraysView(AlignedViewMixin, PairwiseArraysBase):
#     def __init__(
#         self, parent_mapping: PairwiseArraysBase, parent_view: "AnnData", subset_idx
#     ):
#         self.parent_mapping = parent_mapping
#         self._parent = parent_view
#         self.subset_idx = (subset_idx, subset_idx)
#         self._axis = parent_mapping._axis
#         self.dim_names = parent_mapping.dim_names[subset_idx]


# PairwiseArraysBase._view_class = PairwiseArraysView
# PairwiseArraysBase._actual_class = PairwiseArrays



import warnings
from functools import wraps, singledispatch
from typing import Mapping, Any, Sequence, Union, Tuple

import pandas as pd
import numpy as np
from scipy.sparse import spmatrix

from .logging import get_logger

logger = get_logger(__name__)


@singledispatch
def convert_to_dict(obj) -> dict:
    return dict(obj)

@convert_to_dict.register(dict)
def convert_to_dict_dict(obj: dict):
    return obj

@convert_to_dict.register(np.ndarray)
def convert_to_dict_ndarray(obj: np.ndarray):
    if obj.dtype.fields is None:
        raise TypeError(
            "Can only convert np.ndarray with compound dtypes to dict, passed "
            "array had '{}'.".format(obj.dtype)
        )
    return {k: obj[k] for k in obj.dtype.fields.keys()}

@convert_to_dict.register(type(None))
def convert_to_dict_nonetype(obj: None):
    return dict()


def make_index_unique(index: pd.Index, join: str = '-'):
    """Makes the index unique by appending '1', '2', etc.

    The first occurance of a non-unique value is ignored.

    Parameters
    ----------
    join
         The connecting string between name and integer.

    Examples
    --------
    >>> adata1 = sc.AnnData(np.ones((3, 2)), {'obs_names': ['a', 'b', 'c']})
    >>> adata2 = sc.AnnData(np.zeros((3, 2)), {'obs_names': ['d', 'b', 'b']})
    >>> adata = adata1.concatenate(adata2)
    >>> adata.obs_names
    Index(['a', 'b', 'c', 'd', 'b', 'b'], dtype='object')
    >>> adata.obs_names_make_unique()
    >>> adata.obs_names
    Index(['a', 'b', 'c', 'd', 'b-1', 'b-2'], dtype='object')
    """
    if index.is_unique:
        return index
    from collections import defaultdict
    values = index.values
    indices_dup = index.duplicated(keep='first')
    values_dup = values[indices_dup]
    counter = defaultdict(lambda: 0)
    for i, v in enumerate(values_dup):
        counter[v] += 1
        values_dup[i] += join + str(counter[v])
    values[indices_dup] = values_dup
    index = pd.Index(values)
    return index


def warn_names_duplicates(attr: str):
    names = 'Observation' if attr == 'obs' else 'Variable'
    logger.info(
        '{} names are not unique. '
        'To make them unique, call `.{}_names_make_unique`.'
        .format(names, attr))


def warn_no_string_index(names: Sequence[Any]):
    if not isinstance(names[0], str):
        logger.warning(
            'AnnData expects string indices for some functionality, but your first two indices are: {}. '
            .format(names[:2]))


def convert_dictionary_to_structured_array(source: Mapping[str, Sequence[Any]]):
    names = list(source.keys())
    try:  # transform to byte-strings
        cols = [np.asarray(col) if np.array(col[0]).dtype.char not in {'U', 'S'}
                else np.asarray(col).astype('U') for col in source.values()]
    except UnicodeEncodeError:
        raise ValueError(
            'Currently only support ascii strings. Don\'t use "ö" etc. for sample annotation.')

    # if old_index_key not in source:
    #     names.append(new_index_key)
    #     cols.append(np.arange(len(cols[0]) if cols else n_row).astype('U'))
    # else:
    #     names[names.index(old_index_key)] = new_index_key
    #     cols[names.index(old_index_key)] = cols[names.index(old_index_key)].astype('U')
    dtype_list = list(zip(names, [str(c.dtype) for c in cols], [(c.shape[1],) for c in cols]))
    # might be unnecessary
    dtype = np.dtype(dtype_list)

    arr = np.zeros((len(cols[0]),), dtype)
    # here, we do not want to call BoundStructArray.__getitem__
    # but np.ndarray.__getitem__, therefore we avoid the following line
    # arr = np.ndarray.__new__(cls, (len(cols[0]),), dtype)
    for i, name in enumerate(dtype.names):
        arr[name] = np.array(cols[i], dtype=dtype_list[i][1])

    return arr


def deprecated(new_name: str):
    """
    This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used.
    """
    def decorator(func):
        @wraps(func)
        def new_func(*args, **kwargs):
            warnings.simplefilter('always', DeprecationWarning)  # turn off filter
            warnings.warn(
                'Use {0} instead of {1}, {1} will be removed in the future.'
                .format(new_name, func.__name__),
                category=DeprecationWarning,
                stacklevel=2,
            )
            warnings.simplefilter('default', DeprecationWarning)  # reset filter
            return func(*args, **kwargs)
        setattr(new_func, '__deprecated', True)
        return new_func
    return decorator


class DeprecationMixinMeta(type):
    """
    Use this as superclass so deprecated methods and properties
    do not appear in vars(MyClass)/dir(MyClass)
    """
    def __dir__(cls):
        def is_deprecated(attr):
            if isinstance(attr, property):
                attr = attr.fget
            return getattr(attr, '__deprecated', False)

        return [
            item for item in type.__dir__(cls)
            if not is_deprecated(getattr(cls, item, None))
        ]


Index = Union[slice, int, np.int64, np.ndarray, spmatrix]


def get_n_items_idx(idx: Index, l: int):
    if isinstance(idx, np.ndarray) and idx.dtype == bool:
        return idx.sum()
    elif isinstance(idx, slice):
        start = 0 if idx.start is None else idx.start
        stop = l if idx.stop is None else idx.stop
        step = 1 if idx.step is None else idx.step
        return (stop - start) // step
    elif isinstance(idx, (int, np.int_, np.int64, np.int32)):
        return 1
    else:
        return len(idx)


def unpack_index(index: Union[Index, Tuple[Index, Index]]) -> Tuple[Index, Index]:
    # handle indexing with boolean matrices
    if (
        isinstance(index, (spmatrix, np.ndarray))
        and index.ndim == 2
        and index.dtype.kind == 'b'
    ): return index.nonzero()

    if not isinstance(index, tuple):
        return index, slice(None)
    elif len(index) == 2:
        return index
    elif len(index) == 1:
        return index[0], slice(None)
    else:
        raise IndexError('invalid number of indices')



from copy import deepcopy
from functools import reduce, singledispatch
from typing import Any, KeysView, Optional, Sequence, Tuple, NamedTuple

import numpy as np
import pandas as pd
from scipy import sparse

from ..logging import anndata_logger as logger


class ViewArgs(NamedTuple):
    parent: "AnnData"
    attrname: str
    keys: Tuple[str, ...] = ()


class _SetItemMixin:
    def __setitem__(self, idx: Any, value: Any):
        if self._view_args is None:
            super().__setitem__(idx, value)
        else:
            adata_view, attr_name, keys = self._view_args
            logger.warning(
                'Trying to set attribute `.{}` of view, making a copy.'.format(attr_name))
            new = adata_view.copy()
            attr = getattr(new, attr_name)
            container = reduce(lambda d, k: d[k], keys, attr)
            container[idx] = value
            adata_view._init_as_actual(new)


class _ViewMixin(_SetItemMixin):
    def __init__(
        self,
        *args,
        view_args: Tuple['AnnData', str, Tuple[str, ...]] = None,
        **kwargs
    ):
        if view_args is not None:
            view_args = ViewArgs(*view_args)
        self._view_args = view_args
        super().__init__(*args, **kwargs)

    def __deepcopy__(self, memo):
        parent, attrname, keys = self._view_args
        return deepcopy(getattr(parent._adata_ref, attrname))


class ArrayView(_SetItemMixin, np.ndarray):
    def __new__(
        cls,
        input_array: Sequence[Any],
        view_args: Tuple['AnnData', str, Tuple[str, ...]] = None,
    ):
        arr = np.asarray(input_array).view(cls)
        if view_args is not None:
            view_args = ViewArgs(*view_args)
        arr._view_args = view_args
        return arr

    def __array_finalize__(self, obj: Optional[np.ndarray]):
        if obj is not None:
            self._view_args = getattr(obj, '_view_args', None)

    def keys(self) -> KeysView[str]:
        # it's a structured array
        return self.dtype.names

    def copy(self, order: str = 'C') -> np.ndarray:
        # we want a conventional array
        return np.array(self)

    def toarray(self) -> np.ndarray:
        return self.copy()


class SparseCSRView(_ViewMixin, sparse.csr_matrix):
    pass


class SparseCSCView(_ViewMixin, sparse.csc_matrix):
    pass


class DictView(_ViewMixin, dict):
    pass


class DataFrameView(_ViewMixin, pd.DataFrame):
    _metadata = ['_view_args']


@singledispatch
def asview(obj, view_args):
    raise NotImplementedError(
        "No view type has been registered for {}".format(type(obj))
    )

@asview.register(np.ndarray)
def asview_array(array, view_args):
    return ArrayView(array, view_args=view_args)

@asview.register(pd.DataFrame)
def asview_df(df, view_args):
    return DataFrameView(df, view_args=view_args)

@asview.register(sparse.csr_matrix)
def asview_csr(mtx, view_args):
    return SparseCSRView(mtx, view_args=view_args)

@asview.register(sparse.csc_matrix)
def asview_csc(mtx, view_args):
    return SparseCSCView(mtx, view_args=view_args)

@asview.register(dict)
def asview_dict(d, view_args):
    return DictView(d, view_args=view_args)
